package org.postgresql.util;


import java.io.InputStream;
import java.io.IOException;
import java.util.Properties;


/**
 * Provides a location to get properties that the driver should use
 * It looks in the /org/postgresql/psqlProps.properties file first, 
 * which should be set by developers of the PostgreSQL driver. The user
 * can override defaults by having a file in the toplevel of their 
 * classpath called /psqlProps.properties.  Users should not modify the
 * defaults that come with the driver directly.
 */
public class DefaultProperties {

    /**
     * This contains the name of the system default property file used
     * by the postgresql driver.
     */
    private static final String ourProperties ="/org/postgresql/psqlProps.properties";
    /**
     * This contains the name of the user defaults property file, that
     * overrides the system defaults.
     */
    private static final String userProperties="/psqlProps.properties";

    /**
     * Properties variable that contains properties in use.
     */
    private Properties psqlDefaults;

    /**
     * Default constructor.  Should an error occur, properties will be
     * empty, but this will still be usable.
     */
    public DefaultProperties() {
        reload();
    }

    /**
     * Used to hold onto the singleton.  Generally, when I have a singleton,
     * then the default constructor is unavailable.However, this is
     * not the case here.
     */
    private static DefaultProperties self = new DefaultProperties();


    /**
     * Used to get the singleton.  Generally, when I have a singleton,
     * then the default constructor is unavailable.  However, this is
     * not the case here.
     */
    public static  DefaultProperties getDefaultProperties()  {
        return self;
    }

    /**
     * This can be called to reload the properties so that any changes
     * that may have taken place in the properties can be updated.
     */
    public void reload() {
        psqlDefaults = new Properties();
        InputStream is = getClass().getResourceAsStream(ourProperties);
        if (is!=null){
            try {
                psqlDefaults.load(is);
            } catch (IOException e){
                // don't fail, but print the error
                e.printStackTrace();
            }
        }
        is = getClass().getResourceAsStream(userProperties);
        if (is!=null){
            try {
                psqlDefaults.load(is);
            } catch (IOException e){
                // don't fail, but print the error
                e.printStackTrace();
            }
        }
    }
    
    
    /**
     * Accessor to a string property.
     * @param String name of the property
     * @param String default value if the property isn't defined.
     * @return String result
     */
    public String getStringProperty(String name, String defaultValue){
        String value = psqlDefaults.getProperty(name,defaultValue);
        if (value==null){
            return defaultValue;
        } else {
            return value;
        }
    }
    /**
     * Accessor to an int property. If the property value doesn't
     * map properly into the return type, the default value is returned.
     * @param String name of the property
     * @param int default value if the property isn't defined.
     * @return int result
     */
    public int getIntProperty(String name, int defaultValue){
        try {
            String value = psqlDefaults.getProperty(name);
            if (value==null){
                return defaultValue;
            } else {
                return Integer.parseInt(value);
            }
        } catch (NumberFormatException e){
            e.printStackTrace();
            return defaultValue;
        }
    }

    /**
     * Accessor to a long property. If the property value doesn't
     * map properly into the return type, the default value is returned.
     * @param String name of the property
     * @param long default value if the property isn't defined.
     * @return long result
     */
    public long getLongProperty(String name, long defaultValue){
        try {
            String value = psqlDefaults.getProperty(name);
            if (value==null){
                return defaultValue;
            } else {
                return Long.parseLong(value);
            }
        } catch (NumberFormatException e){
            e.printStackTrace();
            return defaultValue;
        }
    }

    /**
     * Accessor to double property. If the property value doesn't
     * map properly into the return type, the default value is returned.
     * @param String name of the property
     * @param double default value if the property isn't defined.
     * @return double result
     */
    public double getDoubleProperty(String name, double defaultValue){
        try {
            String value = psqlDefaults.getProperty(name);
            if (value==null){
                return defaultValue;
            } else {
                return Double.parseDouble(value);
            }
        } catch (NumberFormatException e){
            e.printStackTrace();
            return defaultValue;
        }
    }
    
    /**
     * Accessor to a boolean property.  If the property value doesn't
     * map properly into the return type, the default value is returned.
     * @param String name of the property
     * @param boolean default value if the property isn't defined.
     * @return boolean result
     */
    public boolean getBooleanProperty(String name, boolean defaultValue){
        try {
            String value = psqlDefaults.getProperty(name);
            if (value==null){
                return defaultValue;
            } else {
                return Boolean.valueOf(value).booleanValue();
            }
        } catch (NumberFormatException e){
            e.printStackTrace();
            return defaultValue;
        }
    }

    /**
     * A way to test to make sure the properties that are used are
     * defined in the property files.
     */
    public static void main(String[]argv){
        try {
            DefaultProperties dp = DefaultProperties.getDefaultProperties();
            for (int i=0;i<argv.length;i++){
                String result = dp.getStringProperty(argv[i],"undefined");
                System.out.println("["+argv[i]+"] = " + result);
            }
        } catch (Exception e){
            e.printStackTrace();
        }
    }
    
}
