package org.postgresql.jdbc2;

/* 
 * Author: Ned Wolpert <wolpert@yahoo.com>  
 * For Use By: Anyone under the postgresql license
 */

import javax.sql.PooledConnection;
import javax.sql.ConnectionEvent;
import javax.sql.ConnectionEventListener;
import java.sql.Connection;
import java.sql.Statement;
import java.sql.SQLException;
import java.sql.ResultSet;
import java.sql.PreparedStatement;
import java.sql.CallableStatement;
import java.sql.SQLWarning;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Vector;
import org.postgresql.Field;
import org.postgresql.fastpath.Fastpath;
import org.postgresql.largeobject.LargeObjectManager;
import org.postgresql.util.PSQLException;
import org.postgresql.core.Encoding;

/**
 * This class provides the implementation to the PooledConnection
 * interface for the ConnectionPoolDataSource. It also wrapps the
 * Connection so we can catch errors and exceptions that occured and
 * notify our listeners.
 *
 * TODO: ResultSet, Statement, PreparedStatement and CallableStatement
 * need to be wrapped to catch errors.  Also, need to have a
 * lastTimeUnused() variable so the 'cleaning' thread can auto-close this.
 */
public class PostgresqlPooledConnection implements javax.sql.PooledConnection, java.sql.Connection {
    
    /**
     * This is the real postgresql connection object.
     */
    private org.postgresql.jdbc2.Connection _connection = null;

    /**
     * This is our list of listeners to us.
     */
    private ArrayList connectionEventListeners = new ArrayList();

    /**
     * Constructor allows for the wrapping of connections. Checks to make sure
     * the connection we are wrapping is not null or invalid. If so, we 
     * throw an error.
     */
    public PostgresqlPooledConnection(org.postgresql.jdbc2.Connection connection) throws SQLException{
        if (connection==null){
            throw new SQLException("Null connection tried to be used");
        }
        if (connection.isClosed()){
            throw new SQLException("Connection is closed. " + connection);
        }
        _connection = connection;
    }

    /**
     * Provide method to allow event listeners to add themselves to us
     */
    public synchronized void addConnectionEventListener(ConnectionEventListener listener){
        if (connectionEventListeners.contains(listener)){
            Exception e = new Exception("Already loaded");
            e.printStackTrace();
        } else {
            connectionEventListeners.add(listener);
        }
    }
    
    /**
     * Provide method to allow event listeners to remove themselves from us
     */
    public synchronized void removeConnectionEventListener(ConnectionEventListener listener){        
        connectionEventListeners.remove(listener);
    }

    /**
     * We fake the close here, and notify our listeners instead.  Remember,
     * this is the PooledConnection and the wrapped real Connection.  We
     * are making sure no can really close the connection except our pooling
     * agent through the callRealClose() method.  Why? Because the pooling
     * agent is 100% responsible for this.
     */
    public void close() throws java.sql.SQLException{
        broadcastClosed(new ConnectionEvent(this));
    }

    /**
     * This shouldn't be public, but should only be available to the
     * PostgresqlConnectionPoolDataSource object.  It is in charge of
     * closing the connection for real.
     */
    public void callRealClose() throws java.sql.SQLException {
        _connection.close();
    }

    /**
     * Returns a valid SQL connection... us.  We wrap the real one.
     */
    public java.sql.Connection getConnection() throws java.sql.SQLException{
        return this;
    }

    /**
     * Notify our listeners someone want this connection closed.
     */
    protected void broadcastClosed(ConnectionEvent evt){
        for (Iterator i = connectionEventListeners.iterator();i.hasNext();){
            ConnectionEventListener cel = (ConnectionEventListener)i.next();
            cel.connectionClosed(evt);
        }
    }

    /**
     * Notify our listeners that an error occured.  Currently, we only
     * wrap Connection errors, not resultset or statment errors
     */
    protected void broadcastError(ConnectionEvent evt){
        for (Iterator i = connectionEventListeners.iterator();i.hasNext();){
            ConnectionEventListener cel = (ConnectionEventListener)i.next();
            cel.connectionErrorOccurred(evt);
        }
    }

    // methods here are used by the pooling manager
    private Exception currentException = null;
    public boolean hasException(){
        return (currentException!=null);
    }

    protected Exception getException(){
        return currentException;
    }

    /**
     * Sets us as errored out
     */
    protected void setException(java.sql.SQLException e)  {
        currentException   = e;
        broadcastError(new ConnectionEvent(this,e));
    }

    //
    // Below here are the wrappers of the psql conenction methods.
    // Done this way so we can catch when exceptions occur.
    //
    
    public java.sql.Statement createStatement() throws SQLException{
        try {
            return _connection.createStatement();
        }catch(SQLException e){
            setException(e);
            throw e;
        }
    }


    public java.sql.Statement createStatement(int resultSetType, int resultSetConcurrency) throws SQLException{
        try {
            return _connection.createStatement(resultSetType,resultSetConcurrency);
        }catch(SQLException e){
            setException(e);
            throw e;
        }
    }

    public java.sql.PreparedStatement prepareStatement(String sql) throws SQLException {
        try {
            return _connection.prepareStatement(sql);
        }catch(SQLException e){
            setException(e);
            throw e;
        }
    }

	public java.sql.PreparedStatement prepareStatement(String sql, int resultSetType, int resultSetConcurrency) throws SQLException{
        try {
            return _connection.prepareStatement(sql, resultSetType, resultSetConcurrency);
        }catch(SQLException e){
            setException(e);
            throw e;
        }
    }
    
    public java.sql.CallableStatement prepareCall(String sql) throws SQLException{
        try {
            return _connection.prepareCall(sql);
        }catch(SQLException e){
            setException(e);
            throw e;
        }
    }

    public java.sql.CallableStatement prepareCall(String sql, int resultSetType, int resultSetConcurrency) throws SQLException{
        try {
            return _connection.prepareCall(sql,resultSetType,resultSetConcurrency);
        }catch(SQLException e){
            setException(e);
            throw e;
        }
    }

    public boolean isClosed() throws SQLException{
        try {
            return _connection.isClosed();
        }catch(SQLException e){
            setException(e);
            throw e;
        }
    }

	public java.sql.DatabaseMetaData getMetaData() throws SQLException{
        try {
            return _connection.getMetaData();
        }catch(SQLException e){
            setException(e);
            throw e;
        }
    }

	public java.sql.ResultSet getResultSet(org.postgresql.Connection conn, java.sql.Statement stat, Field[] fields, Vector tuples, String status, int updateCount, int insertOID, boolean binaryCursor) throws SQLException{
        try {
            return _connection.getResultSet( conn,  stat, fields,  tuples,  status,  updateCount,  insertOID,  binaryCursor);
        }catch(SQLException e){
            setException(e);
            throw e;
        }
    }

	public java.util.Map getTypeMap() throws SQLException{
        try {
            return _connection.getTypeMap();
        }catch(SQLException e){
            setException(e);
            throw e;
        }
    }

	public void setTypeMap(java.util.Map map) throws SQLException {
		// new in 7.1
        try {
            _connection.setTypeMap(map);
        }catch(SQLException e){
            setException(e);
            throw e;
        }
	}

    public Object getObject(String type, String value) throws SQLException{
        
        try {
            return _connection.getObject(type,value);
        }catch(SQLException e){
            setException(e);
            throw e;
        }
    }

    // from the org.postgresql.Connection

    public java.sql.ResultSet ExecSQL(String sql) throws SQLException{
        try {
            return _connection.ExecSQL(sql);
        }catch(SQLException e){
            setException(e);
            throw e;
        }
	}

    	public java.sql.ResultSet ExecSQL(String sql, java.sql.Statement stat) throws SQLException{
        try {
            return _connection.ExecSQL(sql,stat);
        }catch(SQLException e){
            setException(e);
            throw e;
        }
	}

	public void setCursorName(String cursor) throws SQLException {
        try {
            _connection.setCursorName(cursor);
        }catch(SQLException e){
            setException(e);
            throw e;
        }
    }

    public String getCursorName() throws SQLException {
        try {
            return _connection.getCursorName();
        }catch(SQLException e){
            setException(e);
            throw e;
        }
    }
    
	public String getURL() throws SQLException{
        try {
            return _connection.getURL();
        } catch (SQLException e){
            setException(e);
            throw e;
        }
	}

    public String getUserName() throws SQLException{
        
        try {
            return _connection.getUserName();
        } catch (SQLException e){
            setException(e);
            throw e;
        }
    }
    
	public Encoding getEncoding() throws SQLException{
        
        try {
            return _connection.getEncoding();
        } catch (SQLException e){
            setException(e);
            throw e;
        }
    }

	public Fastpath getFastpathAPI() throws SQLException{
        try {
            return _connection.getFastpathAPI();
        } catch (SQLException e){
            setException(e);
            throw e;
        }
    }

    public LargeObjectManager getLargeObjectAPI() throws SQLException{
        try {
            return _connection.getLargeObjectAPI();
        } catch (SQLException e){
            setException(e);
            throw e;
        }
    }

    public int putObject(Object o) throws SQLException {
        try {
            return _connection.putObject(o);
        } catch (SQLException e){
            setException(e);
            throw e;
        }
        
    }
	public String nativeSQL(String sql) throws SQLException{
        try {
            return _connection.nativeSQL(sql);
        }catch(SQLException e){
            setException(e);
            throw e;
        }
    }

	public SQLWarning getWarnings() throws SQLException{
        try {
            return _connection.getWarnings();
        }catch(SQLException e){
            setException(e);
            throw e;
        }
    }

	public void clearWarnings() throws SQLException{
        try {
             _connection.clearWarnings();
        }catch(SQLException e){
            setException(e);
            throw e;
        }
    }


	public void setReadOnly(boolean readOnly) throws SQLException{
        try {
            _connection.setReadOnly(readOnly);
        }catch(SQLException e){
            setException(e);
            throw e;
        }
    }

	public boolean isReadOnly() throws SQLException{
        try {
            return _connection.isReadOnly();
        }catch(SQLException e){
            setException(e);
            throw e;
        }
    }

	public void setAutoCommit(boolean autoCommit) throws SQLException{
        try {
            _connection.setAutoCommit(autoCommit);
        }catch(SQLException e){
            setException(e);
            throw e;
        }
    }

	public boolean getAutoCommit() throws SQLException{
        try {
            return _connection.getAutoCommit();
        }catch(SQLException e){
            setException(e);
            throw e;
        }
    }

	public void commit() throws SQLException{
        try {
            _connection.commit();
        }catch(SQLException e){
            setException(e);
            throw e;
        }
    }

	public void rollback() throws SQLException{
        try {
            _connection.rollback();
        }catch(SQLException e){
            setException(e);
            throw e;
        }
    }

	public int getTransactionIsolation() throws SQLException{
        try {
            return _connection.getTransactionIsolation();
        }catch(SQLException e){
            setException(e);
            throw e;
        }
    }

	public void setTransactionIsolation(int level) throws SQLException{
        try {
            _connection.setTransactionIsolation(level);
        }catch(SQLException e){
            setException(e);
            throw e;
        }
    }

	public void setCatalog(String catalog) throws SQLException{
        try {
            _connection.setCatalog(catalog);
        }catch(SQLException e){
            setException(e);
            throw e;
        }
    }
    
	public String getCatalog() throws SQLException{
        try {
            return _connection.getCatalog();
        }catch(SQLException e){
            setException(e);
            throw e;
        }
    }
    public void finalize() throws Throwable {
        callRealClose(); // a real close
    }
    
    public boolean haveMinimumServerVersion(String ver) throws SQLException{
        try {
            return _connection.haveMinimumServerVersion(ver);
        }catch(SQLException e){
            setException(e);
            throw e;
        }
    }

	public boolean haveMinimumCompatibleVersion(String ver) throws SQLException{
        try {
            return _connection.haveMinimumCompatibleVersion(ver);
        }catch(SQLException e){
            setException(e);
            throw e;
        }
    }

	public int getSQLType(int oid) throws SQLException{
        try {
            return _connection.getSQLType(oid);
        }catch(SQLException e){
            setException(e);
            throw e;
        }
    }

	public int getOID(String typeName) throws SQLException{
        try {
            return _connection.getOID(typeName);
        }catch(SQLException e){
            setException(e);
            throw e;
        }
    }

    public String getPGType(int oid) throws SQLException{
        try {
            return _connection.getPGType(oid);
        }catch(SQLException e){
            setException(e);
            throw e;
        }
    }

}
